unit MainForm;

interface

uses NSys, NApp, NForm, NCommand, NConf, NGraphics;

var
  Config: TConf;
  ProfDir: string;

type
  TDateData = record
    Year, Month, Day, YearDay: Integer;
    SolarTerm, HolidayName: string;
  end;

  TCalendar = array [0..41] of TDateData; 

  TColorSchemeStyle = record
    BackgroundColor: Integer;
    Bold: Boolean;
    Color: Integer;
    FrameColor: Integer;
  end;

  TColorScheme = record
    Name: string;
    BackgroundColor: Integer;
    Days: array [0..6] of TColorSchemeStyle;
    Holiday: TColorSchemeStyle;
    Month: TColorSchemeStyle;
    OuterDay: TColorSchemeStyle;
    ThisMonth: TColorSchemeStyle;
    Today: TColorSchemeStyle;
    WeekNumber: TColorSchemeStyle;
  end;

  TNkCalMainForm = class(TForm)
    Calendar: TCalendar;
    ColorScheme: TColorScheme;
    FirstWDay: Integer;
    FontBold: Boolean;
    FontItalic: Boolean;
    FontName: string;
    Font: TFont;
    Month: Integer;
    OldDay: Integer;
    OuterDays: Boolean;
    Timer: TTimer;
    Today: TTime;
    WeekNumber: Boolean;
    Year: Integer;
    YearDay: Boolean;
    SolarTerm: Boolean;
    HolidayName: Boolean;
    Command: TCommand;
    constructor Create; override;
    destructor Destroy; override;
    procedure Closing; override;
    procedure DispatchCommand(Id: Longint); override;
    procedure Paint(const Canvas: TCanvas; Left, Top, Width, Height: Longint);
        override;
    procedure ShowContextMenu(SX, SY: Longint); override;
    procedure UpdateCalendar;
    procedure UpdateToday;
    procedure SetStyle(const Canvas: TCanvas; const Style: TColorSchemeStyle;
        X, Y, W, H, FontSize: Integer);
  published
    procedure cmAbout;
    procedure cmExit;
    procedure cmNextMonth;
    procedure cmNextYear;
    procedure cmPrevMonth;
    procedure cmPrevYear;
    procedure cmShowWeekNumber;
    procedure cmShowYearDay;
    procedure cmShowSolarTerm;
    procedure cmShowHolidayName;
    procedure cmStayOnTop;
    procedure cmThisMonth;
    procedure cmZoom100;
    procedure cmZoom150;
    procedure cmZoom200;
    procedure cmZoom300;
  end;

implementation

uses NLib, NDate, Menu, AboutDialog, Holiday, SolarTerm;

const
  DayName: array [0..6] of string = (
    'Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat', 'Sun'
  );

  ColorSchemes: array [0..1] of TColorScheme = (
    (
      Name: 'default';
      BackgroundColor: $ffffff;
      Days: (
        (BackgroundColor: -1; Color: 0; FrameColor: -1),
        (BackgroundColor: -1; Color: 0; FrameColor: -1),
        (BackgroundColor: -1; Color: 0; FrameColor: -1),
        (BackgroundColor: -1; Color: 0; FrameColor: -1),
        (BackgroundColor: -1; Color: 0; FrameColor: -1),
        (BackgroundColor: -1; Color: $4848f0; FrameColor: -1),
        (BackgroundColor: -1; Color: $f04848; FrameColor: -1)
        );
      Holiday: (BackgroundColor: -1; Color: $ff0000; FrameColor: -1);
      Month: (BackgroundColor: -1; Color: 0; FrameColor: -1);
      Outerday: (BackgroundColor: -1; Color: $e0e0e0; FrameColor: -1);
      ThisMonth: (BackgroundColor: $e6eaf2; FrameColor: -1);
      Today: (BackgroundColor: $e6eaf2; FrameColor: -1);
      WeekNumber: (BackgroundColor: -1; Color: $a0a0a0; FrameColor: -1);
    ),
    (
      Name: 'paper';
      BackgroundColor: $eeeeee;
      Days: (
        (BackgroundColor: -1; Color: 0; FrameColor: -1),
        (BackgroundColor: -1; Color: 0; FrameColor: -1),
        (BackgroundColor: -1; Color: 0; FrameColor: -1),
        (BackgroundColor: -1; Color: 0; FrameColor: -1),
        (BackgroundColor: -1; Color: 0; FrameColor: -1),
        (BackgroundColor: -1; Color: 0; FrameColor: -1),
        (BackgroundColor: -1; Bold: True; Color: $555555; FrameColor: -1)
        );
      Holiday: (BackgroundColor: -1; Bold: True; Color:$555555; FrameColor:-1);
      Month: (BackgroundColor: -1; Color: 0; FrameColor: -1);
      Outerday: (BackgroundColor: -1; Color: $aaaaaa; FrameColor: -1);
      ThisMonth: (BackgroundColor: -1; FrameColor: -1);
      Today: (BackgroundColor: -1; FrameColor: 0);
      WeekNumber: (BackgroundColor: -1; Color: 0; FrameColor: -1);
    )
  );

function GetCalendar(Year, Month, FirstWeekDay: Integer): TCalendar;
  function todatedata(const Date: TDate): TDateData;
  begin
    Result.Year := Date.Year;
    Result.Month := Date.Month;
    Result.Day := Date.Day;
    Result.YearDay := Date.YearDay;
    Result.SolarTerm := SolarTermName(Date);
    Result.HolidayName := HolidayName(Date);
  end;
var
  d, i: Integer;
  wd1, wde, py, pm, pdm, prev, ny, nm: Integer;
begin
  for i := 0 to High(Result) do
    with Result[i] do
    begin
      Year := 0;
      Month := 0;
      Day := 0;
      YearDay := 0;
      HolidayName := '';
    end;

  if Month = 0 then
    Exit;

  wd1 := Date(Year, Month, 1).WeekDay;
  wde := Date(Year, Month, DaysInMonth(Year, Month)).WeekDay;
  i := 0;

  // previous month days
  prev := (wd1 - FirstWeekDay + 7) mod 7;
  if Month = 1 then
  begin
    py := Year-1;
    pm := 12;
  end
  else begin
    py := Year;
    pm := Month-1;
  end;
  pdm := DaysInMonth(py, pm);
  for d := pdm-prev+1 to pdm do
  begin
    Result[i] := todatedata(Date(py, pm, d));
    Inc(i);
  end;

  // this month days
  for d := 1 to DaysInMonth(Year, Month) do
  begin
    Result[i] := todatedata(Date(Year, Month, d));
    Inc(i);
  end;

  // next month days
  if Month = 12 then
  begin
    ny := Year+1;
    nm := 1;
  end
  else begin
    ny := Year;
    nm := Month+1;
  end;
  for d := 1 to (FirstWeekday - wde + 6) mod 7 do
  begin
    Result[i] := todatedata(Date(ny, nm, d));
    Inc(i);
  end;
end;

constructor TNkCalMainForm.Create;
var
  arg1, arg2, colorscheme: string;
  i: Integer;
begin
  inherited;

  cmZoom150;

  Command := TCommand.Create([
    cmAbout,
    cmExit,
    cmNextMonth,
    cmNextYear,
    cmPrevMonth,
    cmPrevYear,
    cmShowWeekNumber,
    cmShowYearDay,
    cmShowSolarTerm,
    cmShowHolidayName,
    cmStayOnTop,
    cmThisMonth,
    cmZoom100,
    cmZoom150,
    cmZoom200,
    cmZoom300
    ]);

  LoadFormPos(Self, Config);
  FirstWDay := Config.GetInt('firstweekday');
  FontName := Config.GetStr('font');
  arg1 := Config.GetStr('font', 1);
  arg2 := Config.GetStr('font', 2);
  FontBold := (arg1 = 'bold') or (arg2 = 'bold');
  FontItalic := (arg1 = 'italic') or (arg2 = 'italic');
  OuterDays := Config.GetBool('outerdays');
  WeekNumber := Config.GetBool('weeknumber');
  YearDay := Config.GetBool('yearday');
  HolidayName := Config.GetBool('holidayname');
  SolarTerm := Config.GetBool('solarterm');
  colorscheme := Config.GetStr('colorscheme');

  Self.ColorScheme := ColorSchemes[0];
  for i := 0 to High(ColorSchemes) do
    if ColorSchemes[i].Name = colorscheme then
    begin
      Self.ColorScheme := ColorSchemes[i];
      Break;
    end;

  SetBackground(Self.ColorScheme.BackgroundColor);

  UpdateToday;
  cmThisMonth;
  Timer := TTimer.Create(Self, UpdateToday, 1000);
  Timer.Start;

  Show;
end;

destructor TNkCalMainForm.Destroy;
begin
  Timer.Free;
  Command.Free;
end;

procedure TNkCalMainForm.ShowContextMenu(SX, SY: Longint);
begin
  with TPopupMenu.Create do
    try
      Command.AddMenuItems(Handle, MenuItems);
      CheckItem(Command.Id(cmStayOnTop), StayOnTop);
      CheckItem(Command.Id(cmShowWeekNumber), WeekNumber);
      CheckItem(Command.Id(cmShowYearDay), YearDay);
      CheckItem(Command.Id(cmShowHolidayName), HolidayName);
      CheckItem(Command.Id(cmShowSolarTerm), SolarTerm);
      Popup(Self.Handle, SX, SY);
    finally
      Free;
    end;
end;

procedure TNkCalMainForm.DispatchCommand(Id: Longint);
begin
  Command.DispatchCommand(Id);
end;

procedure TNkCalMainForm.SetStyle(const Canvas: TCanvas;
    const Style: TColorSchemeStyle; X, Y, W, H, FontSize: Integer);
var
  newfont: TFont;
begin
  with Style do
  begin
    if BackgroundColor > -1 then
      Canvas.FillRect(X, Y, W, H, BackgroundColor);
    if FrameColor > -1 then
      Canvas.DrawRect(X, Y, W, H, FrameColor);
    newfont := TFont.Create(FontName, -FontSize, FontBold or Bold, FontItalic);
    Canvas.SetFont(newfont);
    Font.Free;
    Font := newfont;
    Canvas.SetTextColor(Color);
  end;
end;

procedure TNkCalMainform.Paint(const Canvas: TCanvas;
    Left, Top, Width, Height: Longint);
var
  d: TDateData;
  cw, ch, h, dw, x, y, i, j, wd: Integer;
  fontsize, daynamefontsize, yeardayfontsize: Integer;
  style: TColorSchemeStyle;
  s: WideString;
begin
  GetClientSize(cw, ch);
  h := ch div 8;
  fontsize := h * 90 div 100;
  daynamefontsize := fontsize * 3 div 4;
  yeardayfontsize := fontsize div 3;

  Canvas.BeginDraw;

  Canvas.SetTextAlign(taCenter);

  // year and month
  if (Year = Today.Year) and (Month = Today.Mon) then
    SetStyle(Canvas, ColorScheme.ThisMonth, 0, 0, cw, h, fontsize)
  else
    SetStyle(Canvas, ColorScheme.Month, 0, 0, cw, h, fontsize);
  Canvas.DrawText(cw div 2, (h-fontsize) div 2,
      Format('{}-{:02}', [Year, Month]));

  // day name
  y := h + (h-daynamefontsize) div 2;
  if WeekNumber then
  begin
    dw := cw div 8;
    x := dw div 2;
    SetStyle(Canvas, ColorScheme.WeekNumber, 0, y, dw, h, daynamefontsize);
    Canvas.DrawText(x, y, 'W');
    Inc(x, dw);
  end
  else begin
    dw := cw div 7;
    x := dw div 2;
  end;
  for i := 0 to 6 do
  begin
    j := (FirstWDay+i) mod 7;
    SetStyle(Canvas, ColorScheme.Days[j], x-dw div 2, y, dw, h,
        daynamefontsize);
    Canvas.DrawText(x, y, DayName[j]);
    Inc(x, dw);
  end;

  // days
  wd := FirstWDay;
  y := h*2;
  for i := 0 to 5 do
  begin
    if WeekNumber then
      x := dw
    else
      x := 0;
    for j := 0 to 6 do
    begin
      d := Calendar[i*7+j];
      if d.Day = 0 then
        Break;

      if WeekNumber and (wd = 0) then
      begin
        SetStyle(Canvas, ColorScheme.WeekNumber, 0, y, dw, h, daynamefontsize);
        Canvas.DrawText(dw div 2, y + (h-daynamefontsize) div 2,
            IStr(Date(d.Year, d.Month, d.Day).ISOWeekNumber));
      end;

      if d.Month = Month then
      begin
        if d.HolidayName <> '' then
          style := ColorScheme.Holiday
        else
          style := ColorScheme.Days[(FirstWDay+j) mod 7];
      end else
        style := ColorScheme.OuterDay;
      if ((d.Year=Year) and (d.Month=Month) and (d.Day=Today.MDay) and 
           (Year=Today.Year) and (Month=Today.Mon)) or
         ((d.Year=Today.Year) and (d.Month=Today.Mon) and
           (d.Day=Today.MDay) and OuterDays) then
      begin
        style.FrameColor := ColorScheme.Today.FrameColor;
        style.BackgroundColor := ColorScheme.Today.BackgroundColor;
      end;
      SetStyle(Canvas, style, x, y, dw, h, fontsize);

      if (d.Month = Month) or OuterDays then
        if YearDay or HolidayName or SolarTerm then begin
          Canvas.DrawText(x + dw div 2, y, IStr(d.Day));
          if d.Month = Month then
          begin
            SetStyle(Canvas, ColorScheme.WeekNumber, x, y, dw, h,
                yeardayfontsize);
            s := '';
            if YearDay then
            begin
              if s <> '' then
                s := s + ' ';
              s := s + IStr(d.YearDay) + '-' +
                  IStr(365+Ord(IsLeap(d.Year))-d.YearDay);
            end;
            if SolarTerm then
            begin
              if s <> '' then
                s := s + ' ';
              s := s + d.SolarTerm;
            end;
            if HolidayName and (d.HolidayName <> '') then
            begin
              if s <> '' then
                s := s + ' ';
              s := s + d.HolidayName;
            end;
            Canvas.DrawText(x + dw div 2, y + h - yeardayfontsize, s);
          end;
        end else
          Canvas.DrawText(x + dw div 2, y + (h-fontsize) div 2, IStr(d.Day));
      wd := (wd+1) mod 7;
      Inc(x, dw);
    end;
    Inc(y, h);
  end;
  
  Canvas.EndDraw;
end;

procedure TNkCalMainForm.Closing;
var
  position: TConf;
begin
  if not Config.GetBool('saveposition') then
    Exit;

  position := TConf.Create;
  try
    SaveFormPos(Self, position);
    try
      position.SaveToFile(JoinPath([ProfDir, 'position']));
    except
      on E: EOSError do App.Error(E.Path+#10#10+E.Msg);
    end;
  finally
    position.Free;
  end;
end;

procedure TNkCalMainForm.UpdateToday;
begin
  OldDay := Today.MDay;
  Today := LocalTime(Time);
  if OldDay <> Today.MDay then 
  begin
    Year := Today.Year;
    Month := Today.Mon; 
    UpdateCalendar;
  end;
end;

procedure TNkCalMainForm.UpdateCalendar;
begin
  Calendar := GetCalendar(Year, Month, FirstWDay);
  SetText(Format('{}-{:02} - {}', [Year, Month, App.Name]));
  Invalidate;
end;

{ Command Methods }

procedure TNkCalMainForm.cmStayOnTop;
begin
  SetStayOnTop(not StayOnTop);
end;

procedure TNkCalMainForm.cmPrevMonth;
begin
  if Month = 1 then
  begin
    Month := 12;
    Dec(Year);
  end else
    Dec(Month);
  UpdateCalendar;
end;

procedure TNkCalMainForm.cmNextMonth;
begin
  if Month = 12 then
  begin
     Month := 1;
     Inc(Year);
  end else
     Inc(Month);
  UpdateCalendar;
end;

procedure TNkCalMainForm.cmThisMonth;
begin
  Year := Today.Year;
  Month := Today.Mon; 
  UpdateCalendar;
end;

procedure TNkCalMainForm.cmNextYear;
begin
  Inc(Year);
  UpdateCalendar;
end;

procedure TNkCalMainForm.cmPrevYear;
begin
  Dec(Year);
  UpdateCalendar;
end;

procedure TNkCalMainForm.cmZoom100;
begin
  SetClientSize(320, 240);
end;

procedure TNkCalMainForm.cmZoom150;
begin
  SetClientSize(480, 360);
end;

procedure TNkCalMainForm.cmZoom200;
begin
  SetClientSize(640, 480);
end;

procedure TNkCalMainForm.cmZoom300;
begin
  SetClientSize(960, 720);
end;

procedure TNkCalMainForm.cmShowWeekNumber;
begin
  WeekNumber := not WeekNumber;
  Invalidate;
end;

procedure TNkCalMainForm.cmShowYearDay;
begin
  YearDay := not YearDay;
  Invalidate;
end;

procedure TNkCalMainForm.cmShowSolarTerm;
begin
  SolarTerm := not SolarTerm;
  Invalidate;
end;

procedure TNkCalMainForm.cmShowHolidayName;
begin
  HolidayName := not HolidayName;
  Invalidate;
end;

procedure TNkCalMainForm.cmAbout;
begin
  with TAboutDialog.Create do
    try
      ProfDir := MainForm.ProfDir;
      ShowModal(Self);
    finally
      Free;
    end;
end;

procedure TNkCalMainForm.cmExit;
begin
  Close;
end;

end.
